#!/usr/bin/env python
"""
genairics.jobs contains all the logics for submitting genairics pipelines as jobs
to for example a qsub system
"""
from plumbum import local, SshMachine
from luigi.contrib.ssh import RemoteTarget
from luigi.util import inherits
import luigi, os, time

# Tasks
from genairics import logger, config

class QueuJob(luigi.Task):
    """
    Submits a pipeline as a qsub job.
    #TODO give option to wait until job finishes and print job stdout and stderr to console
    """
    job = luigi.TaskParameter(description = 'the pipeline that will be submitted as a qsub job')
    resources = luigi.DictParameter(
        default = {'walltime':'20:00:00','nodes':config.nodes, 'ppn':config.threads},
        description = 'the resources that will be asked by the qsub job'
    )
    remote = luigi.Parameter('', description='provide ssh config remote name, if job is to be submitted remotely')

    def output(self):
        return (
            luigi.LocalTarget('{}/../results/{}/plumbing'.format(self.job.datadir,self.job.project))
            if not self.remote else
            RemoteTarget('{}/../results/{}/plumbing'.format(self.job.datadir,self.job.project),host=self.remote),
            luigi.LocalTarget('{}/../results/{}/plumbing/completed_{}'.format(self.job.datadir,self.job.project,self.task_family))
            if not self.remote else
            RemoteTarget('{}/../results/{}/plumbing/completed_{}'.format(self.job.datadir,self.job.project,self.task_family),host=self.remote)
        )

    def run(self):
        machine = SshMachine(self.remote) if self.remote else local
        jobvariables = ['GENAIRICS_ENV_ARGS={}'.format(self.job.task_family),'SET_LUIGI_FRIENDLY=']
        for n in self.job.get_param_names():
            if self.job.__getattribute__(n): jobvariables.append('{}={}'.format(n,self.job.__getattribute__(n)))
        qsub = machine['qsub'][
            '-l','walltime={}'.format(self.resources['walltime']),
            '-l','nodes={}:ppn={}'.format(self.resources['nodes'],self.resources['ppn']),
        ]
        qsubID = qsub('-v',','.join(jobvariables),machine['genairics'])
        logger.warning('qsub job submitted with ID %s. Waiting for result dir to be ready...',qsubID)
        # Wait until pipeline project result dir exists
        while not self.output()[0].exists():
            time.sleep(60)
        # When it is created during the job's execution, touch job submission completion file
        machine['touch'](self.output()[1].path)
        if self.remote: machine.close()

def DockerJob(args):
    """
    Submits a pipeline to be run in the genairics docker container.
    """
    raise NotImplementedError
    #docker = local['docker']
    #docker run -v ~/resources:/resources -v ~/data:/data -v ~/results:/results --env-file ~/.BASESPACE_API beukueb/genairics RNAseq -h



#!/usr/bin/env python
#PBS -N RNAseqPipeline
#PBS -l nodes=1:ppn=16
#PBS -l walltime=72:00:00
#PBS -m be
"""
Full pipeline starting from BaseSpace fastq project
"""
from datetime import datetime, timedelta
import luigi, os, tempfile, pathlib, glob
from luigi.contrib.external_program import ExternalProgramTask
from luigi.util import inherits, requires
from plumbum import local, colors
import pandas as pd
import logging

# matplotlib => setup for exporting svg figures only
import matplotlib
matplotlib.use('SVG')
import matplotlib.pyplot as plt

## Tasks
from genairics import logger, config, gscripts, setupProject, setupSequencedSample
from genairics.datasources import BaseSpaceSource, mergeFASTQs
from genairics.resources import resourcedir, STARandRSEMindex

@inherits(mergeFASTQs)
class qualityCheck(luigi.Task):
    """
    Runs fastqc on all samples and makes an overall summary
    """
    def requires(self):
        return self.clone_parent()
        
    def output(self):
        return (
            luigi.LocalTarget('{}/{}/plumbing/completed_{}'.format(self.resultsdir,self.project,self.task_family)),
            luigi.LocalTarget('{}/{}/QCresults'.format(self.resultsdir,self.project)),
            luigi.LocalTarget('{}/{}/summaries/qcsummary.csv'.format(self.resultsdir,self.project))
        )

    def run(self):
        import zipfile
        from io import TextIOWrapper
        
        local[gscripts % 'qualitycheck.sh'](self.project, self.datadir)
        qclines = []
        for fqcfile in glob.glob(self.output()[1].path+'/*.zip'):
            zf = zipfile.ZipFile(fqcfile)
            with zf.open(fqcfile[fqcfile.rindex('/')+1:-4]+'/summary.txt') as f:
                ft = TextIOWrapper(f)
                summ = pd.read_csv(TextIOWrapper(f),sep='\t',header=None)
                qclines.append(summ[2].ix[0]+'\t'+'\t'.join(list(summ[0]))+'\n')
        with self.output()[2].open('w') as outfile:
            outfile.writelines(['\t'+'\t'.join(list(summ[1]))+'\n']+qclines)
        pathlib.Path(self.output()[0].path).touch()

# per sample subtasks
## STAR aligning
@inherits(STARandRSEMindex)
class STARconfig(luigi.Config):
    """
    Reference: https://github.com/alexdobin/STAR
    """
    readFilesCommand = luigi.Parameter(
        default='zcat',
        description='STAR readFilesCommand parameter'
    )
    outSAMtype = luigi.Parameter(
        default='BAM SortedByCoordinate',
        description='STAR outSAMtype parameter (can contain more than one argument separated by 1 space)'
    )
    quantMode = luigi.Parameter(
        default='TranscriptomeSAM GeneCounts',
        description='STAR quantMode parameter (can contain more than one argument separated by 1 space)'
    )    

@inherits(STARconfig)
@inherits(setupSequencedSample)
class STARsample(luigi.Task):
    """
    Task that does the STAR alignment

    Currently fq's not moved first to tmp dir TODO
    Previous implementation that did that in the bash script:
    #Prepare workdir
    if [ "$PBS_JOBID" ]; then
    cd $TMPDIR
    if [ -d fastqs ]; then
	# if quality check ran previously on same node, fastqs will already be present
	mkdir alignmentResults
    else
	mkdir {fastqs,alignmentResults}
	cp $datadir/$project/*.fastq.gz fastqs/
    fi
    outdir=$TMPDIR/alignmentResults
    else
    cd $datadir/../results/$project/
    mkdir alignmentResults
    outdir=$datadir/../results/$project/alignmentResults
    fi

    at the end tmp results were moved to final destination:
    if [ "$PBS_JOBID" ]; then
    mv $TMPDIR/alignmentResults $datadir/../results/${project}/alignmentResults${suffix}
    fi
    """
    def requires(self):
        return self.clone(setupSequencedSample)

    def output(self):
        return luigi.LocalTarget('{}/.completed_{}'.format(self.outfileDir,self.task_family))
        
    def run(self):
        stdout = local['STAR'](
            '--runThreadN', config.threads,
            '--genomeDir', resourcedir+'/ensembl/{species}/release-{release}/transcriptome_index'.format(
                species=self.genome,release=self.release),
            '--readFilesIn', self.infile1, *((self.infile2,) if self.infile2 else ()), 
	    '--readFilesCommand', self.readFilesCommand,
	    '--outFileNamePrefix', os.path.join(self.input().path,'./'),
	    '--outSAMtype', *self.outSAMtype.split(' '),
	    '--quantMode', *self.quantMode.split(' ')
        )
        if stdout: logger.info('%s output:\n%s',self.task_family,stdout)

        # Check point
        pathlib.Path(self.output().path).touch()

## RSEM counting
@inherits(STARconfig)
class RSEMconfig(luigi.Config):
    """
    Reference: http://deweylab.biostat.wisc.edu/rsem/README.html
    http://deweylab.biostat.wisc.edu/rsem/rsem-calculate-expression.html

    Important documentation:
    --forward-prob <double> Probability of generating a read from the
    forward strand of a transcript. Set to 1 for a strand-specific
    protocol where all (upstream) reads are derived from the forward
    strand, 0 for a strand-specific protocol where all (upstream) read
    are derived from the reverse strand, or 0.5 for a non-strand-specific
    protocol.
    """
    forwardprob = luigi.FloatParameter(
        default=0.5,
        description='stranded seguencing [0 for illumina stranded], or non stranded [0.5]'
    )
    
@inherits(RSEMconfig)
@inherits(STARsample)
class RSEMsample(luigi.Task):
    def requires(self):
        return [
            self.clone(setupSequencedSample),
            self.clone(STARsample)
        ]

    def output(self):
        return luigi.LocalTarget('{}/.completed_{}'.format(self.outfileDir,self.task_family))

    def run(self):
        stdout = local['rsem-calculate-expression'](
            '-p', config.threads, '--alignments',
            *(('--paired-end',) if self.infile2 else ()),
            '--forward-prob', self.forwardprob,
            os.path.join(self.input()[0].path,'Aligned.toTranscriptome.out.bam'),
            resourcedir+'/ensembl/{species}/release-{release}/transcriptome_index/{species}'.format(
                species=self.genome,release=self.release),
            os.path.join(self.input()[0].path,os.path.basename(self.outfileDir))
        )
        if stdout: logger.info('%s output:\n%s',self.task_family,stdout)
        
        # Check point
        pathlib.Path(self.output().path).touch()

# the sample pipeline can inherit and clone the sample subtasks directly
@inherits(RSEMsample)
class processTranscriptomicSampleTask(luigi.Task):
    """
    This wrappers makes sure all the individuel sample tasks get run.
    Each task should be idempotent to avoid issues.
    """
    def output(self):
        return luigi.LocalTarget('{}/.completed_{}'.format(self.outfileDir,self.task_family))
    
    def run(self):
        self.clone(setupSequencedSample).run()
        self.clone(STARsample).run()
        self.clone(RSEMsample).run()
        pathlib.Path(self.output().path).touch()

# the all samples pipeline needs to inherit the sample pipeline configs
@inherits(mergeFASTQs)
@inherits(RSEMconfig)    
class processTranscriptomicSamples(luigi.Task):
    """
    Process transciptomic samples for RNAseq with STAR aligner
    """
    suffix = luigi.Parameter(default='',description='use when preparing for xenome filtering')

    def requires(self):
        return {
            'fastqs':self.clone(mergeFASTQs)
        }

    def output(self):
        return (
            luigi.LocalTarget('{}/{}/plumbing/completed_{}'.format(self.resultsdir,self.project,self.task_family)),
            luigi.LocalTarget('{}/{}/sampleResults'.format(self.resultsdir,self.project)),
        )

    def run(self):
        # Make output directory
        if not self.output()[1].exists(): os.mkdir(self.output()[1].path)

        # Run the sample subtasks
        for fastqfile in glob.glob(os.path.join(
                self.datadir,self.project,
                '*_R1.fastq.gz' if self.pairedEnd else '*.fastq.gz')
        ):
            sample = os.path.basename(fastqfile).replace('.fastq.gz','')
            processTranscriptomicSampleTask( #OPTIONAL future implement with yield
                infile1 = fastqfile,
                infile2 = fastqfile.replace('_R1.','_R2.') if self.pairedEnd else '',
                outfileDir = os.path.join(self.output()[1].path,sample), #optionally in future first to temp location
                **{k:self.param_kwargs[k] for k in RSEMconfig.get_param_names()}
            ).run()
        
        # Check point
        pathlib.Path(self.output()[0].path).touch()
        
@requires(processTranscriptomicSamples)
class mergeAlignResults(luigi.Task):
    """
    Merge the align and count results
    """

    def output(self):
        return (
            luigi.LocalTarget('{}/{}/plumbing/completed_{}'.format(self.resultsdir,self.project,self.task_family)),
            luigi.LocalTarget('{}/{}/summaries/STARcounts.csv'.format(self.resultsdir,self.project)),
            luigi.LocalTarget('{}/{}/summaries/RSEMcounts.csv'.format(self.resultsdir,self.project))
        )

    def run(self):
        #Process STAR counts
        amb = []
        counts = []
        amb_annot = counts_annot = None
        samples = []
        for dir in glob.glob(os.path.join(self.input()[1].path,'*')):
            f = open(os.path.join(dir,'ReadsPerGene.out.tab'))
            f = f.readlines()
            amb.append([int(l.split()[1]) for l in f[:4]])
            if not amb_annot: amb_annot = [l.split()[0] for l in f[:4]]
            f = f[4:]
            if not counts_annot: counts_annot = [l.split()[0] for l in f]
            else:
                assert counts_annot == [l.split()[0] for l in f]
            counts.append([int(l.split()[1]) for l in f])
            samples.append(dir[dir.rindex('/')+1:])
        # Alignment summary file
        counts_df = pd.DataFrame(counts,columns=counts_annot,index=samples).transpose()
        counts_df.to_csv(self.output()[1].path)
    
        # Process RSEM counts
        counts = {}
        samples = []
        for gfile in glob.glob(os.path.join(self.input()[1].path,'*/*.genes.results')):
            sdf = pd.read_table(gfile,index_col=0)
            counts[gfile[gfile.rindex('/')+1:-14]] = sdf.expected_count

        # Counts summary file
        counts_df = pd.DataFrame(counts)
        counts_df.to_csv(self.output()[2].path)
        
        # Check point
        pathlib.Path(self.output()[0].path).touch()

@requires(mergeAlignResults)
class PCAplotCounts(luigi.Task):
    sampleAnnotator = luigi.Parameter(
        default = '_',
        description = 'the string that separates annotation parts in the sample filename'
    )
    annotatorRelevant = luigi.IntParameter(
        default = 1,
        description = 'the number of relevant annotation parts in the sample filename'
    )
    countsFilter = luigi.FloatParameter(
        default = 1,
        description = 'the minimum average number (over all samples) of reads required for a gene for further analysis'
    )
    PCAcomponents = luigi.IntParameter(
        default = 3,
        description = 'number of PCA components to calculate'
    )
    
    def output(self):
        return [
            luigi.LocalTarget('{}/{}/summaries/PCAplot.svg'.format(self.resultsdir,self.project)),
            luigi.LocalTarget('{}/{}/summaries/PCAvars.svg'.format(self.resultsdir,self.project)),
            luigi.LocalTarget('{}/{}/summaries/PCAcomps.csv'.format(self.resultsdir,self.project)),
        ]

    def run(self):
        import matplotlib
        matplotlib.use('svg')
        import matplotlib.pyplot as plt
        import seaborn as sns
        import numpy as np
        from sklearn import preprocessing
        from sklearn.decomposition import PCA as sklearnPCA
        counts = pd.read_csv(self.input()[2].path, index_col = 'gene_id')
        # Filter counts
        beforeFiltering = len(counts)
        counts = counts[counts.sum(axis=1) >= self.countsFilter * len(counts.columns)]
        logger.info('Remaining %s of %s after filtering',len(counts),beforeFiltering)
        # Normalize
        quantile_transformer = preprocessing.QuantileTransformer(
            output_distribution = 'normal',
            n_quantiles = 100,
            random_state = 0
        )
        X_norm = quantile_transformer.fit_transform(counts.T.as_matrix())
        #X_scaled = preprocessing.scale(counts.as_matrix())
        #X_norm = (X - X.min())/(X.max() - X.min())
        # Transform counts
        X_tran = np.log(X_norm + abs(X_norm.min()) + 1)
        pca = sklearnPCA(n_components=self.PCAcomponents)
        transformed = pd.DataFrame(
            pca.fit_transform(X_tran), index = counts.columns,
            columns = ['PC{}'.format(pc+1) for pc in range(self.PCAcomponents)]
        )
        transformed['annotation'] = transformed.T.apply(
            lambda x: ' '.join(x.name.split(self.sampleAnnotator)[:self.annotatorRelevant])
        )
        pcaComponents = pd.DataFrame(pca.components_.T,index=counts.index,columns=transformed.columns[:-1])
        pcaComponents.to_csv(self.output()[2].path)
        # Make figures
        ## PC plot
        fig = sns.lmplot('PC1', 'PC2', data = transformed, hue = 'annotation', fit_reg = False, scatter_kws={'s':50})
        fig.savefig(self.output()[0].path)
        ## PC variance plot
        fix, ax = plt.subplots()
        sns.barplot(np.arange(self.PCAcomponents), pca.explained_variance_ratio_, ax=ax)
        fig.savefig(self.output()[1].path)
        
@requires(mergeAlignResults)
class diffexpTask(luigi.Task):
    design = luigi.Parameter(default='',
                             description='model design for differential expression analysis')
    
    def output(self):
        return (
            luigi.LocalTarget('{}/{}/plumbing/completed_{}'.format(self.resultsdir,self.project,self.task_family)),
            luigi.LocalTarget('{}/{}/summaries/DEexpression.csv'.format(self.resultsdir,self.project))
        )

    def run(self):
        if not self.metafile:
            samples = glob.glob('{}/{}/sampleResults/*'.format(self.resultsdir,self.project))
            samples = pd.DataFrame(
                {'bam_location':samples,
                 'alignment_dir_size':[local['du']['-sh'](s).split('\t')[0] for s in samples]},
                index = [os.path.basename(s) for s in samples]
            )
            metafile = '{}/{}/metadata/samples.csv'.format(self.resultsdir,self.project)
            samples.to_csv(metafile)
            msg = colors.red | '''
                metafile needs to be provided to run DE analysis
                a template file has been generated for you ({})
                adjust file to match your design, add the above file path
                as input "metafile" for the pipeline and rerun
                '''.format(metafile)
            logger.error(msg)
            raise Exception()
        with local.env(R_MODULE="SET"):
            local['bash'][
                '-l','-c', ' '.join(
                    ['Rscript', gscripts % 'simpleDEvoom.R',
                     self.project, self.datadir, self.metafile, self.design]
                )]()
        pathlib.Path(self.output()[0].path).touch()

@inherits(BaseSpaceSource)
@inherits(PCAplotCounts)
@inherits(diffexpTask)
class RNAseq(luigi.WrapperTask):
    def requires(self):
        yield self.clone(setupProject)
        yield self.clone(BaseSpaceSource)
        yield self.clone(mergeFASTQs)
        yield self.clone(qualityCheck)
        yield self.clone(processTranscriptomicSamples)
        yield self.clone(mergeAlignResults)
        yield self.clone(PCAplotCounts)
        if self.design: yield self.clone(diffexpTask)



# -*- coding: utf-8 -*-
"""Datasources

Connecting genairics pipelines to different (raw) data sources.

Todo:
    * https://support.illumina.com/sequencing/sequencing_software/bcl2fastq-conversion-software.html
"""

import luigi, logging, os, sys, itertools, glob, pathlib
from luigi.util import inherits, requires
from plumbum import local
from genairics import config, setupProject
from genairics.resources import RetrieveGenome

@inherits(setupProject)
class BaseSpaceSource(luigi.Task):
    """
    Uses the BaseSpace API from Illumina for downloading.
    It takes the project name and downloads the fastq files.

    The task is completed when a datadir folder exists with the project name
    so if you do not need to download it, just manually put the data in the datadir
    with the project name.
    """
    NSQrun = luigi.Parameter('',description='sequencing run project name')
    basespace_API_file = luigi.Parameter(
        config.basespaceAPIfile,
        description = 'file that contains your basespace API token. Should only be readable by your user',
        significant = False
    )

    def requires(self):
        return self.clone_parent()

    def output(self):
        return luigi.LocalTarget('{}/{}'.format(self.datadir,self.project))
    
    def run(self):
        import requests, tempfile
        logger = logging.getLogger(__package__)

        # Check if NSQrun is set, otherwise set to project name
        if not self.NSQrun:
            self.NSQrun = self.project
            logger.warning('NSQrun was not provided, assuming same as project %s' % self.project)

        # Load basespace token
        if os.path.exists(self.basespace_API_file):
            BASESPACE_API_TOKEN = open(self.basespace_API_file).read().strip().replace('BASESPACE_API_TOKEN=','')
        elif 'BASESPACE_API_TOKEN' in os.environ:
            BASESPACE_API_TOKEN = os.environ['BASESPACE_API_TOKEN']
        else:
            logger.error('BASESPACE_API_TOKEN not in file or environment')
            raise Exception()

        # Find the project ID
        request = 'http://api.basespace.illumina.com/v1pre3/users/current/projects?access_token=%s&limit=1000' % (BASESPACE_API_TOKEN,)
        r = requests.get(request)
        projectName = False
        for project in r.json()['Response']['Items']:
            if project['Name'] == self.NSQrun:
                (projectName, projectID) = (project['Name'], project['Id'])
                break
    
        if not projectName:
            logger.error('Project {} not found on BaseSpace'.format(self.NSQrun))
            raise Exception()

        # Prepare temp dir for downloading
        outtempdir = tempfile.mkdtemp(prefix=self.datadir+'/',suffix='/')

        # Find project sample IDs (max 1000)
        request = 'http://api.basespace.illumina.com/v1pre3/projects/%s/samples?access_token=%s&limit=1000' % (projectID, BASESPACE_API_TOKEN)
        r = requests.get(request)
        for sample in r.json()['Response']['Items']:
            (sampleName, sampleID) = (sample['Name'], sample['Id'])
            logger.info('Retrieving '+sampleName)
            sampleDir = os.path.join(outtempdir, sampleName)
            os.mkdir(sampleDir)
            sample_request = 'http://api.basespace.illumina.com/v1pre3/samples/%s/files?access_token=%s' % (sampleID, BASESPACE_API_TOKEN)
            sample_request = requests.get(sample_request)
            for sampleFile in sample_request.json()['Response']['Items']:
                filePath = os.path.join(sampleDir, sampleFile['Path'])
                logger.info('Path: '+filePath)
                if not os.path.isfile(filePath):
                    file_request = 'http://api.basespace.illumina.com/%s/content?access_token=%s' % (sampleFile['Href'], BASESPACE_API_TOKEN)
                    file_request = requests.get(file_request, stream=True)
                    with open(filePath,'wb') as outfile:
                        for chunk in file_request.iter_content(chunk_size=512):
                            outfile.write(chunk)
                #check downloas
                if sampleFile['Size'] != os.path.getsize(filePath):
                    logger.error('size of local file and BaseSpace file do not match')
                    raise Exception()

        # Rename tempdir to final project name dir
        os.rename(outtempdir,self.output().path)

@requires(BaseSpaceSource)
class mergeFASTQs(luigi.Task):
    """
    Merge fastqs if one sample contains more than one fastq
    """
    dirstructure = luigi.Parameter(
        default='multidir',
        description='dirstructure of project datat directory: onedir (one file/sample) or multidir (one dir/sample)'
    )
    pairedEnd = luigi.BoolParameter(
        default=False,
        description='paired end sequencing reads'
    )
    removeOriginalFQs = luigi.BoolParameter(
        default=False,
        description='remove the folder with the original, unmerged FASTQ files'
    )
    
    def output(self):
        return (
            luigi.LocalTarget('{}/{}/.completed_{}'.format(self.datadir,self.project,self.task_family)),
            luigi.LocalTarget('{}/{}/.{}.log'.format(self.datadir,self.project,self.task_family))
        )

    def run(self):
        if self.dirstructure == 'multidir' and not self.output()[0].exists():
            outdir = '{}/{}'.format(self.datadir,self.project)
            oridir = '{}/{}_original_FASTQs'.format(self.datadir,self.project)
            os.rename(outdir,oridir)
            os.mkdir(outdir)
            dirsFASTQs = local['ls'](oridir).split()
            for d in dirsFASTQs:
                (local['ls'] >> (self.output()[1].path))('-lh',os.path.join(oridir,d))
                if self.pairedEnd:
                    (local['cat'] > os.path.join(outdir,d+'_R1.fastq.gz'))(
                        *glob.glob(os.path.join(oridir,d,'*_R1_*.fastq.gz'))
                    )
                    (local['cat'] > os.path.join(outdir,d+'_R2.fastq.gz'))(
                        *glob.glob(os.path.join(oridir,d,'*_R2_*.fastq.gz'))
                    )
                else:
                    (local['cat'] > os.path.join(outdir,d+'.fastq.gz'))(
                        *glob.glob(os.path.join(oridir,d,'*.fastq.gz'))
                    )
            if self.removeOriginalFQs:
                import shutil
                shutil.rmtree(oridir)
                logger.warning('%s removed',oridir)
        # complete file and log file get touched even if no execution is necessary
        pathlib.Path(self.output()[0].path).touch()
        pathlib.Path(self.output()[1].path).touch()

def groupfilelines(iterable, n=4, fillvalue=None):
    """
    read n files per iteration
    default n = 4 => read 1 fastq entry at a time

    >>> for lines in groupfilelines(f):
    ...     break
    """
    from itertools import zip_longest
    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=fillvalue)

@requires(mergeFASTQs)
class SubsampleProject(luigi.Task):
    """
    Takes a project datadir and subsamples the FASTQs to either an exact number of
    reads or a percentage of the original files
    """
    amount = luigi.FloatParameter(description='amount of reads to subsample')
    percentage = luigi.BoolParameter(
        False,
        description='amount of reads will be treated as the approximate percentage of reads of the sample'
    )
    seed = luigi.IntParameter(0,'seed to use for random selection')
    subsuffix = luigi.Parameter(
        default = '_subsampled',
        description = 'suffix to use for new subsampled data directory'
    )
    
    def output(self):
        return luigi.LocalTarget('{}/{}_subsampled'.format(self.datadir,self.project))

    def run(self):
        import gzip, random
        random.seed(self.seed)
        oridir = '{}/{}'.format(self.datadir,self.project)
        subdir = '{}/{}{}'.format(self.datadir,self.project,self.subsuffix)
        outdir = '{}_subsampling'.format(subdir)
        os.mkdir(outdir)
        for fqfile in glob.glob(
                os.path.join(
                    oridir,
                    '*._R1.fastq.gz' if self.pairedEnd else '*.fastq.gz'
                )):
            # Determine which reads to subsample
            reads = 0
            with gzip.open(fqfile) as gf1:
                for lines in groupfilelines(gf1): reads += 1
                gf1.seek(0)
                if self.percentage: self.amount = int(reads*self.amount)
                selectedReads = random.sample(range(reads),self.amount)
                with gzip.open(os.path.join(outdir,os.path.basename(fqfile)), 'wb') as gfout:
                    for i,lines in enumerate(groupfilelines(gf1)):
                        if i in selectedReads:
                            gfout.writelines(lines)
            if self.pairedEnd:
                fqfile2 = fqfile.replace('_R1.fastq.gz','_R2.fastq.gz')
                with gzip.open(fqfile2) as gf2, gzip.open(os.path.join(outdir,os.path.basename(fqfile2)), 'wb') as gfout:
                    for i,lines in enumerate(groupfilelines(gf2)):
                        if i in selectedReads:
                            gfout.writelines(lines)
        # renaming temporary outdir to final subdir
        os.rename(outdir,subdir)

@inherits(setupProject)
class ENAsource(luigi.Task):
    """
    Downloads fastq's from given ENA accession number
    """
    ENAaccession = luigi.Parameter('',description='sequencing run project name')

@inherits(setupProject)
@inherits(RetrieveGenome)
class ENAtestSource(luigi.Task):
    """
    Downloads fastq's from given ENA accession number
    """
    def requires(self):
        return [
            self.clone(setupProject),
            self.clone(RetrieveGenome)
        ]

    def run(self):
        import requests
        r = requests.get('https://www.ebi.ac.uk/ena/data/taxonomy/v1/taxon/scientific-name/{}'.format(
            genome.replace('_','%20'))
        )
        taxID = r.json()[0]['taxId']
        r = requests.get('https://www.ebi.ac.uk/ena/data/view/Taxon:{}&portal=read_experiment;display=xml'.format(taxID))
        #https://www.ebi.ac.uk/ena/data/warehouse/search?query=%22cell_line=%22IMR-32%22%22&domain=sample
        #https://www.ebi.ac.uk/ena/data/warehouse/search?query=%22geo_accession=%22GSE37599%22%22&domain=study
        raise NotImplementedError
    
@inherits(setupProject)
@inherits(RetrieveGenome)
class SimulatedSource(luigi.Task):
    def requires(self):
        return [
            self.clone(setupProject),
            self.clone(RetrieveGenome)
        ]

    def run(self):
        import gffutils
        #transform func needed for ensembl gtf => see gffutils docs examples
        def transform_func(x):
            if 'transcript_id' in x.attributes:
                x.attributes['transcript_id'][0] += '_transcript'
            return x
        db = gffutils.create_db(
            glob.glob(os.path.join(self.input()[1].path,'annotation/*.gtf'))[0],':memory:',
            id_spec={'gene': 'gene_id', 'transcript': "transcript_id"},
            merge_strategy="create_unique",
            transform=transform_func,
            keep_order=True
        )
        transcripts = db.features_of_type('transcript')
        #TODO work in progress
        #get transcripts -> to fasta file -> then R polyester for simulation
        raise NotImplementedError



#!/usr/bin/env python
"""
ChIP sequencing pipeline starting from BaseSpace fastq project

References:
 - http://jvanheld.github.io/cisreg_course/chip-seq/practical/chip-seq.html
 - http://crazyhottommy.blogspot.be/2015/06/chip-seq-analysis-part1.html
"""
from datetime import datetime, timedelta
import luigi, os, tempfile, pathlib, glob
from luigi.contrib.external_program import ExternalProgramTask
from luigi.util import inherits, requires
from plumbum import local, colors
import pandas as pd
import logging

## Tasks
from genairics import config, logger, gscripts, setupProject
from genairics.datasources import BaseSpaceSource, mergeFASTQs
from genairics.resources import resourcedir, RetrieveGenome, Bowtie2Index, RetrieveBlacklist
from genairics.RNAseq import qualityCheck

### ChIP specific Task
class cutadaptConfig(luigi.Config):
    """
    Info: http://cutadapt.readthedocs.io/en/stable/guide.html
    """
    adapter = luigi.Parameter(
        default = "GATCGGAAGAGCACACGTCTGAACTCCAGTCACCGATGTATCTCGTATGC",
        description = "cutadapt adapter to trim"
    )
    errorRate = luigi.FloatParameter(
        default = 0.1,
        description = "allowed error rate => errors #/length mapping"
    )
    pairedEnd = luigi.BoolParameter(
        default=False,
        description='paired end sequencing reads (NOT IMPLEMENTED YET)'
    )

@inherits(cutadaptConfig)
class TrimFilterSample(luigi.Task):
    infile = luigi.Parameter()
    outfileDir = luigi.Parameter()

    def output(self):
        return luigi.LocalTarget(os.path.join(self.outfileDir,'trimmed.fq.gz'))
        
    def run(self):
        stdout = local['cutadapt'](
            '--cores', config.threads,
            '-a', self.adapter,
            '-e', self.errorRate,
            '-o', self.output().path,
            self.infile
        )
        if stdout: logger.info(stdout)

#subsampleTask => subsampling naar 30 miljoen indien meer

@inherits(TrimFilterSample)
@inherits(Bowtie2Index)
class Bowtie2MapSample(luigi.Task):
    def requires(self):
        return {
            'sample': self.clone(TrimFilterSample),
            'index': self.clone(Bowtie2Index)
            }
        
    def output(self):
        return luigi.LocalTarget(os.path.join(self.outfileDir,'alignment.bam'))

    def run(self):
        # run bowtie2 and store as bam file with mapping quality already filtered to mapQ 4
        stdout = (local['bowtie2'][
            '-p', config.threads,
            '-x', os.path.join(self.input()['index'][0].path,self.genome),
            '-U', self.input()['sample'].path #-U -> unpaired, TODO to start using paired will be with -1 and -2
        ] | local['samtools']['view', '-q', 4, '-Sbh', '-'] > self.output().path)()
        if stdout: logger.info(stdout)

@requires(Bowtie2MapSample)
class SamProcessSample(luigi.Task):
    def output(self):
        return [
            luigi.LocalTarget(os.path.join(self.outfileDir,'alignment_sorted.bam')),
            luigi.LocalTarget(os.path.join(self.outfileDir,'alignment_sorted.bai')),
            luigi.LocalTarget(os.path.join(self.outfileDir,'flagstatsummary.txt')),
            luigi.LocalTarget(os.path.join(self.outfileDir,'idxstats.txt'))
        ]

    def run(self):
        # sort output
        stdout = local['samtools']('sort', self.input().path, '-o', self.output()[0].path)

        # index output
        stdout += local['samtools']('index', self.output()[0].path)

        # replace unsorted bam with empty stub
        with self.input().open('wb') as f:
            f.truncate()
    
        # TODO samstat on bamfile
        #http://samstat.sourceforge.net/

        # stats: flagstat and per chromosome read stats
        stdout += (local['samtools']['flagstat', self.output()[0].path] > self.output()[2].path)()
        stdout += (local['samtools']['idxstats', self.output()[0].path] > self.output()[3].path)()

        # log
        if stdout: logger.info(stdout)

@requires(SamProcessSample)
class MakeSampleGenomeBrowserTrack(luigi.Task):
    """
    genome browser track file
    """
    def output(self):
        return luigi.LocalTarget(os.path.join(self.outfileDir,'alignment_coverage.bw'))

    def run(self):
        stdout = local['bamCoverage']('-b', self.input()[0].path, '–outFileFormat', 'bigwig', '-o', self.output().path)

# the sample pipeline can inherit and clone the subtasks directly
@inherits(MakeSampleGenomeBrowserTrack)
class processGenomicSampleTask(luigi.WrapperTask):
    def run(self):
        self.clone(TrimFilterSample).run()
        self.clone(Bowtie2MapSample).run()
        self.clone(SamProcessSample).run()
        self.clone(MakeSampleGenomeBrowserTrack).run()

# the all samples pipeline needs to inherit the sample pipeline configs
@inherits(mergeFASTQs)
@inherits(cutadaptConfig)    
@inherits(Bowtie2Index)
class processGenomicSamples(luigi.Task):
    """
    Process genomic samples (can be used for ChIP, ATAC, variant calling)
    """
    def requires(self):
        return {
            'fastqs': self.clone(mergeFASTQs)
        }

    def output(self):
        return (
            luigi.LocalTarget('{}/{}/plumbing/completed_{}'.format(self.resultsdir,self.project,self.task_family)),
            luigi.LocalTarget('{}/{}/alignmentResults'.format(self.resultsdir,self.project)),
        )

    def run(self):
        # Make output directory
        if not self.output()[1].exists(): os.mkdir(self.output()[1].path)

        # Run the sample subtasks
        for fastqfile in glob.glob(os.path.join(self.datadir,self.project,'*.fastq.gz')):
            sample = os.path.basename(fastqfile).replace('.fastq.gz','')
            processGenomicSampleTask( #OPTIONAL future implement with yield
                infile = fastqfile,
                outfileDir = os.path.join(self.output()[1].path,sample+'/'), #optionally in future first to temp location
                **{k:self.param_kwargs[k] for k in cutadaptConfig.get_param_names()}
            ).run()
        
        # Check point
        pathlib.Path(self.output()[0].path).touch()

@requires(processGenomicSamples)
class PeakCallingChIPsamples(luigi.Task):
    """
    performs the peak calling with macs2
    tries to match every ChIPseq sample with an input sample,
    if not working as expected you will have to change the filenames

    an input filename needs to contain 'input' or whatever you set as inputFileMarker
    a matching sample file, needs to contain everything of the input filename, before
    the 'input'/inputFileMarker section

    e.g. inputfile => IMR32_input.fq.gz
    e.g. matching file => IMR32_H3K27ac.fq.gz, as it also contains 'IMR32_'
    """
    inputFileMarker = luigi.Parameter(
        default='input',
        description='should be in filename of every input sample'
    )
    
    def run(self):
        inputfiles = {i for i in glob.glob(os.path.join(self.input()[0].path,'*{}*'.format(self.inputFileMarker)))}
        inputmatchmap = {os.path.basename(i).split(self.inputFileMarker)[0]:i for i in inputfiles}
        if len(inputfiles) != len(inputmatchmap):
            logger.error('some inputfiles have non unique prefix before input marker "%s" (%s)',
                         self.inputFileMarker,inputfiles)
            raise Exception()
        for sample in glob.glob(os.path.join(self.input()[0].path,'*')):
            if sample not in inputfiles:
                foundMarker = False
                for marker in inputmatchmap:
                    if os.basename(sample).startswith(marker):
                        foundMarker = marker
                        break
                if not foundMarker:
                    logger.warning('no matching input file found for %s',os.path.basename(sample))
                else:
                    #macs2
                    stdout = local['macs2'](
                        'callpeak', '-t', os.path.join(sample,'alignment_sorted.bam'),
                        '-c', os.path.join(inputmatchmap[foundMarker],'alignment_sorted.bam'),
                        '--outdir', sample, '-n', 'diff_peaks', '-q', 0.05, '-g', 'hs', '--bdg'
                    )
                    if stdout: logger.info(stdout)

                    # homer
                    stdout = local['makeTagDirectory'](
                        os.path.join(sample,'homer_tags'),
                        os.path.join(sample,'alignment_sorted.bam')
                    )
                    stdout = local['makeTagDirectory'](
                        #TODO good place in the input align dir so does not compute twice in case of reuse
                        os.path.join(sample,'homer_tags_input'),
                        os.path.join(inputmatchmap[foundMarker],'alignment_sorted.bam')
                    )
                    stdout += local['findPeaks'](
                        os.path.join(sample,'homer_tags'),
                        '-style', 'factor',
                        '-o', os.path.join(sample,'homer_peaks_found'),
                        '-i', os.path.join(sample,'homer_tags_input')
                    )
                    stdout += (local['pos2bed.pl'][os.path.join(sample,'homer_peaks_found')]
                               > os.path.join(sample,'homer_peaks.bed'))()
                    if stdout: logger.info(stdout)


#class ClusterBamFiles(luigi.Task):
    #DeepTools for clustering of bam files
    #multiBamSummary bins --bamfiles sam_1/CLBGA_INPUT_F_sorted.bam sam_2/CLBGA_TBX2_F_sorted.bam sam_6/SKNAS_INPUT_B_sorted.bam sam_7/SKNAS_TBX2_B_sorted.bam sam_8/SKNAS_H3K27ac_B_sorted.bam -out multiBamSummary_bam_Run296.npz --labels CLBGA_INPUT CLBGA_TBX2 SKNAS_INPUT SKNAS_TBX2 SKNAS_H3K27ac
    #plotCorrelation --corData multiBamSummary_bam_Run296.npz --plotFile correlation_peaks.pdf --outFileCorMatrix correlation_peaks_matrix.txt --whatToPlot heatmap --corMethod pearson --plotNumbers --removeOutliers

@inherits(BaseSpaceSource)
@inherits(PeakCallingChIPsamples)
class ChIPseq(luigi.WrapperTask):
    def requires(self):
        yield self.clone(setupProject)
        yield self.clone(BaseSpaceSource)
        yield self.clone(mergeFASTQs)
        yield self.clone(qualityCheck)
        yield self.clone(processGenomicSamples)
        yield self.clone(PeakCallingChIPsamples)



#!/usr/bin/env python
"""
Tasks here prepare resources that are required for some
pipeline tasks, and are generally available from the
resources directory as specified by $GAX_RESOURCES or
in genairics config file 'resourcedir' option
"""
from datetime import datetime, timedelta
import luigi, os, sys, tempfile, pathlib, glob
from luigi.util import inherits, requires
from plumbum import local, colors
import logging

from genairics import config, gscripts

resourcedir = config.resourcedir
logresources = logging.Logger('genairics.resources')

#Set cksum program according to platform
if sys.platform == 'linux':
    cksum = local['sum']
elif sys.platform == 'darwin':
    cksum = local['cksum']['-o',1]
else: raise Exception('{} not supported. Run from docker container.'.format(sys.platform))

def requestFiles(urlDir,fileregex,outdir):
    """
    Download a set of files that match fileregex from urlDir
    checksums are expected to be generated by the "sum" command,
    as is the case for the ftp.ensembl.org repository
    """
    import requests, re
    link = re.compile(r'href="(.+?)"')
    if fileregex: fileregex = re.compile(fileregex)
    if not urlDir.endswith('/'): urlDir+='/'
    # Download checksums if present
    checksums = requests.get(urlDir+'CHECKSUMS')
    if checksums:
        from io import StringIO
        import pandas as pd
        checksums = pd.read_table(StringIO(checksums.text),sep='\s+',
                                  names=['checksum', 'octets', 'name'],index_col='name')
    else: checksums = None
    # Retrieve index
    r = requests.get(urlDir)
    for l in link.finditer(r.text):
        l = l.groups()[0]
        if not fileregex or fileregex.match(l):
            file_request = requests.get(urlDir+l, stream=True)
            with open(os.path.join(outdir,l),'wb') as outfile:
                for chunk in file_request.iter_content(chunk_size=512):
                    outfile.write(chunk)
            if checksums is not None:
                csum = int(cksum(os.path.join(outdir,l)).split()[0])
                if csum != checksums.checksum[l]:
                    logresources.warning('%s checksum did not match url location %s',csum,urlDir)

# Tasks
class InstallDependencies(luigi.Task):
    """
    Installs the genairics dependency programs
    """
    def run(self):
        local[gscripts % 'genairics_dependencies.sh']
    
class RetrieveGenome(luigi.Task):
    """
    Prepares the genome
    """
    genome = luigi.Parameter(default='homo_sapiens',description="genome species name")
    release = luigi.IntParameter(default=91,
                                 description=
"""ensembl release number of genome
for homo_sapiens, use the following :
 91 => GRCh38 aka hg38;
 75 => GRCh37 aka hg19
"""
    )

    def output(self):
        return luigi.LocalTarget(
            os.path.join(resourcedir,'ensembl',self.genome,'release-{}'.format(self.release))
        )

    def run(self):
        #Make temp dir for data
        local['mkdir']('-p',self.output().path+'_retrieving/dna')
        local['mkdir']('-p',self.output().path+'_retrieving/annotation')
        requestFiles(
            'http://ftp.ensembl.org/pub/release-{release}/fasta/{species}/dna/'.format(
                species=self.genome, release=self.release),
            r'.+.dna.chromosome.+.fa.gz',
            self.output().path+'_retrieving/dna'
        )
        requestFiles(
            'http://ftp.ensembl.org/pub/release-{release}/gtf/{species}/'.format(
                species=self.genome, release=self.release),
            r'.+.{release}.gtf.gz'.format(release=self.release),
            self.output().path+'_retrieving/annotation'
        )
        #Unzip all files
        local['gunzip'](*glob.glob(self.output().path+'_retrieving/*/*.gz'))
        #Rename temp dir to final/expected output dir
        os.rename(self.output().path+'_retrieving',self.output().path)

@requires(RetrieveGenome)
class RetrieveBlacklist(luigi.Task):
    """
    Info on blacklists: https://sites.google.com/site/anshulkundaje/projects/blacklists

    Available release:
    91 => GRCh38
    75 => GRCh37 (latest release of hg19)
    """
    availableBlacklists = {
        ('homo_sapiens', 91): "http://mitra.stanford.edu/kundaje/akundaje/release/blacklists/hg38-human/hg38.blacklist.bed.gz",
        ('homo_sapiens', 75):
        "http://mitra.stanford.edu/kundaje/akundaje/release/blacklists/hg19-human/wgEncodeHg19ConsensusSignalArtifactRegions.bed.gz"
    }
    
    def output(self):
        return luigi.LocalTarget(os.path.join(self.input().path,'blacklist.bed.gz'))

    def run(self):
        if (self.genome, self.release) not in self.availableBlacklists:
            raise NotImplementedError("blacklist for %s release %s is not available yet within genairics" % (self.genome,self.release))
        stdout = local['wget'](self.availableBlacklists[(self.genome, self.release)], '-O', self.output().path)
        logresources.info(stdout)

@requires(RetrieveGenome)
class Bowtie2Index(luigi.Task):
    """
    Index that can be used by bowtie2 aligner for ChIPseq, ATACseq, variant calling
    """
    def output(self):
        return (
            luigi.LocalTarget(
                os.path.join(resourcedir,'ensembl/{species}/release-{release}/genome_index'.format(
                    species=self.genome,release=self.release))
            ),
            luigi.LocalTarget(
                os.path.join(resourcedir,'ensembl/{species}/release-{release}/genome_index/build_completed'.format(
                    species=self.genome,release=self.release))
            )
        )
    
    def run(self):
        genomeDir = self.input().path
        os.mkdir(self.output()[0].path)
        stdout = local['bowtie2-build'](
            '--threads', config.threads,
            ','.join(glob.glob(os.path.join(genomeDir,'dna')+'/*.fa*')),
            os.path.join(self.output()[0].path, self.genome)
        )
        logresources.info(stdout)
        pathlib.Path(self.output()[1].path).touch()        

@inherits(RetrieveGenome)
class STARandRSEMindex(luigi.Task):
    """
    Index that can be used by both STAR aligner and RSEM counter for transcriptomics
    """
    def requires(self):
        return self.clone_parent()
    
    def output(self):
        return (
            luigi.LocalTarget(
                os.path.join(resourcedir,'ensembl/{species}/release-{release}/transcriptome_index'.format(
                    species=self.genome,release=self.release))
            ),
            luigi.LocalTarget(
                os.path.join(resourcedir,'ensembl/{species}/release-{release}/transcriptome_index/build_completed'.format(
                    species=self.genome,release=self.release))
            )
        )
    
    def run(self):
        genomeDir = self.input().path
        os.mkdir(self.output()[0].path)
        stdout = local[gscripts % 'buildRSEMindex.sh'](
            *glob.glob(os.path.join(genomeDir,'annotation')+'/*.gtf'),
            config.threads,
            ','.join(glob.glob(os.path.join(genomeDir,'dna')+'/*.fa*')),
            os.path.join(self.output()[0].path, self.genome)
        )
        logresources.info(stdout)
        pathlib.Path(self.output()[1].path).touch()

@inherits(RetrieveGenome)
class STARindex(luigi.Task):
    """
    Index that can be used by STAR aligner for genome mapping 
    (for genomic variant discovery, ChiPseq, ATACseq, )

    NOT YET TESTED, for the moment use STARandRSEMindex
    """
    sjdbOverhang = luigi.IntParameter(default=100, description='library prep read length. default of 100 should generally be ok')
    
    def requires(self):
        return self.clone_parent()
    
    def output(self):
        return (
            luigi.LocalTarget(
                os.path.join(resourcedir,'ensembl/{species}/release-{release}/genome_index_overhang{overhang}'.format(
                    species=self.genome,release=self.release,overhang=self.sjdbOverhang))
            ),
            luigi.LocalTarget(
                os.path.join(resourcedir,'ensembl/{species}/release-{release}/genome_index_overhang{overhang}/build_completed'.format(
                    species=self.genome,release=self.release,overhang=self.sjdbOverhang))
            )
        )
    
    def run(self):
        genomeDir = self.input().path
        os.mkdir(self.output()[0].path)
        stdout = local['STAR'](
            '--runThreadN', config.threads,
            '--runMode', 'genomeGenerate',
            '--genomeDir', os.path.join(self.output()[0].path, self.genome),
            '--genomeFastaFiles', ','.join(glob.glob(os.path.join(genomeDir,'dna')+'/*.fa*')),
            '--sjdbOverhang', self.sjdbOverhang
            
        )
        #TODO STAR Log.out will be produced should be moved to genome index dir
        logresources.info(stdout)
        pathlib.Path(self.output()[1].path).touch()



#!/usr/bin/env python
"""
ATAC sequencing pipeline starting from BaseSpace fastq project
"""

from datetime import datetime, timedelta
import luigi, os, tempfile, pathlib, glob
from luigi.contrib.external_program import ExternalProgramTask
from luigi.util import inherits, requires
from plumbum import local, colors
import pandas as pd
import logging

## Tasks
from genairics import config, logger, gscripts, setupProject
from genairics.datasources import BaseSpaceSource, mergeFASTQs
from genairics.resources import resourcedir, STARandRSEMindex, RetrieveBlacklist
from genairics.RNAseq import qualityCheck

### ATAC specific Tasks
class alignSTARconfig(luigi.Config):
    """
    Contains global STAR parameters that should be configurable by the end user.
    STAR parameters that are set automatically reside within the alignATACsampleTask only
    including parameters that are sample specific.
    """
    runThreadN = luigi.IntParameter(config.threads,description="processors that STAR will use")
    runMode = luigi.Parameter("alignReads",description="STAR run mode")
    readFilesCommand = luigi.Parameter("zcat",description="command for decompressing fq file")
    outSAMtype = luigi.Parameter("BAM SortedByCoordinate")
    outFilterMultimapNmax = luigi.IntParameter(
        default=1,
        description="how many mappings/read allowed; 1 to exclude multimapping reads")
    alignIntronMax = luigi.IntParameter(1)
    outWigType = luigi.Parameter("bedGraph")
    outWigNorm = luigi.Parameter("RPM")
    
@inherits(alignSTARconfig)
class alignATACsampleTask(luigi.Task):
    """
    The task to process one sample.
    Intended to be called from a project task that contains all samples needed to be processed.
    """
    genomeDir = luigi.Parameter(description='genome dir')
    readFilesIn = luigi.Parameter(description='fastqfile(s)')
    outFileNamePrefix = luigi.Parameter(description='result destination')

    def output(self):
        return luigi.LocalTarget(self.outFileNamePrefix)

    def run(self):
        os.mkdir(self.outFileNamePrefix)
        stdout = local['STAR'](
            '--runThreadN', self.runThreadN,
            '--runMode', self.runMode,
            '--genomeDir', self.genomeDir,
            '--readFilesIn', self.readFilesIn,
            '--readFilesCommand', self.readFilesCommand,
	    '--outFileNamePrefix', self.outFileNamePrefix,
	    '--outSAMtype', self.outSAMtype.split()[0], self.outSAMtype.split()[1],
	    '--alignIntronMax', self.alignIntronMax,
	    '--outWigType', self.outWigType,
            '--outWigNorm', self.outWigNorm
        )
        logger.info('%s output:\n%s',self.task_family,stdout)
    
@inherits(mergeFASTQs)
@inherits(alignSTARconfig)    
@inherits(STARandRSEMindex)
class alignATACsamplesTask(luigi.Task):
    """
    Align reads to genome with STAR
    TODO pairedEnd processing not implemented yet
    """
    pairedEnd = luigi.BoolParameter(default=False,
                                    description='paired end sequencing reads')
    
    def requires(self):
        return {
            'genome':self.clone(STARandRSEMindex), #OPTIONAL use genome index only instead of the RSEM build transcriptome index
            'fastqs':self.clone(mergeFASTQs)
        }

    def output(self):
        return (
            luigi.LocalTarget('{}/{}/plumbing/completed_{}'.format(self.resultsdir,self.project,self.task_family)),
            luigi.LocalTarget('{}/{}/alignmentResults'.format(self.resultsdir,self.project)),
        )

    def run(self):
        if self.pairedEnd: raise NotImplementedError('paired end not yet implemented')
        
        # Make output directory
        if not self.output()[1].exists(): os.mkdir(self.output()[1].path)

        # Run the sample subtasks
        for fastqfile in glob.glob(os.path.join(self.datadir,self.project,'*.fastq.gz')):
            sample = os.path.basename(fastqfile).replace('.fastq.gz','')
            alignATACsampleTask( #OPTIONAL future implement with yield
                genomeDir=self.input()['genome'][0].path,
                readFilesIn=fastqfile,
                outFileNamePrefix=os.path.join(self.output()[1].path,sample+'/'), #optionally in future first to temp location
                **{k:self.param_kwargs[k] for k in alignSTARconfig.get_param_names()}
            ).run()
        
        # Check point
        pathlib.Path(self.output()[0].path).touch()

@requires(alignATACsamplesTask)
class SamBedFilteringTask(luigi.Task):
    """
    Filtering mapped reads on quality and optionally on blacklisted genomic regions 
    (https://sites.google.com/site/anshulkundaje/projects/blacklists)
    Blacklisting currently only for human data
    """
    filterBlacklist = luigi.BoolParameter(True,description="if human genome, filter blacklisted regions")
    samtoolsViewQ = luigi.IntParameter(4,description="samtools view -q mapping quality parameter for filtering")
    
    def output(self):
        return (
            luigi.LocalTarget('{}/{}/plumbing/completed_{}'.format(self.resultsdir,self.project,self.task_family)),
            self.input()[1] # forward inherited alignment dir
        )

    def run(self):
        for sampleFile in glob.glob(os.path.join(self.input()[1].path,'*')):
            sample = os.path.basename(sampleFile)
            
            # Filtering mapped reads
            stdout = local['samtools'](
                'view', '-b', '-q', self.samtoolsViewQ, '-@', 2,
                '-o', os.path.join(sampleFile,"Aligned.sortedByCoord.minMQ4.bam"),
	        os.path.join(sampleFile,"Aligned.sortedByCoord.out.bam")
            )
            logger.info(stdout)

            #Filtering blacklisted genomic regions
            if self.filterBlacklist and self.genome in {'homo_sapiens'}:
                blacklist = RetrieveBlacklist(genome=self.genome,release=self.release)
                if not blacklist.complete(): blacklist.run()
                (local['bedtools'][
                    'intersect', '-v', '-abam',
                    os.path.join(sampleFile,"Aligned.sortedByCoord.minMQ4.bam"),
                    '-b', blacklist.output().path] > os.path.join(sampleFile,"Filtered.sortedByCoord.minMQ4.bam"))()
                logger.info("blacklist filtering finished")
            else:
                os.rename(
                    os.path.join(sampleFile,"Aligned.sortedByCoord.minMQ4.bam"),
                    os.path.join(sampleFile,"Filtered.sortedByCoord.minMQ4.bam")
                )
                logger.info("without blacklist filtering")

            # Indexing final filtered file
            stdout = local['samtools']('index', os.path.join(sampleFile,"Filtered.sortedByCoord.minMQ4.bam"))
            logger.info(stdout)
            
        # Check point
        pathlib.Path(self.output()[0].path).touch()

@requires(SamBedFilteringTask)
class PeakCallingTask(luigi.Task):
    """
    MACS2 peak calling

    Work in progress -> still decide how to normalize for bw file
    --normalizeUsingRPKM
    --normalizeTo1x 2451960000  
    """
    callSummits = luigi.BoolParameter(default=False,description="lets MACS2 also call subpeaks")
    normalizeTo1x = luigi.IntParameter(
        default=0,
        description="""
        MACS2 normalization option. If not provided will default to --normalizeUsingRPKM.
        Else int needs to be provided of genome size to pass to --normalizeTo1x,
        e.g. 2451960000 for human genome.
        """
    )
    extsize = luigi.IntParameter(
        default=200,
        description="""
        MACS2 extsize option. If O, --nomodel will not be included.
        """
    )
    peakQ = luigi.FloatParameter(
        default=.05,
        description="""
        MACS2 peak calling q option. Cutoff for peak detection.
        """
    )
    
    def output(self):
        return (
            luigi.LocalTarget('{}/{}/plumbing/completed_{}'.format(self.resultsdir,self.project,self.task_family)),
            self.input()[1] # forward inherited alignment dir
        )

    def run(self):
        for sampleFile in glob.glob(os.path.join(self.input()[1].path,'*')):
            sample = os.path.basename(sampleFile)
            with local.env(PYTHONPATH=''):
                stdout = local['macs2'](
                    'callpeak', '-t',
                    os.path.join(sampleFile,"Filtered.sortedByCoord.minMQ4.bam"),
                    '-n', os.path.join(sampleFile,sample),
                    *(('--nomodel','--extsize',self.extsize) if self.extsize else ()),
                    '--nolambda', '-q', self.peakQ,
                    '--keep-dup', 'auto',
                    *(('--call-summits',) if self.callSummits else ())
                )
            if stdout: logger.info(stdout)
            with local.env(PYTHONPATH=''):
                stdout = local['bamCoverage'](
                    '-p', str(config.threads),
                    *(('--normalizeTo1x',self.normalizeTo1x) if self.normalizeTo1x else ('--normalizeUsingRPKM',)),
                    #'--extendReads', #TODO make possible for both single/paired end
		    '-b', os.path.join(sampleFile,"Filtered.sortedByCoord.minMQ4.bam"),
		    '-o', os.path.join(sampleFile,"Filtered.sortedByCoord.minMQ4.bam")[:-3]+'coverage.bw' 
                )
            if stdout: logger.info(stdout)

        # Check point
        pathlib.Path(self.output()[0].path).touch()
    
@inherits(BaseSpaceSource)
@inherits(PeakCallingTask)
class ATACseq(luigi.WrapperTask):
    def requires(self):
        yield self.clone(setupProject)
        yield self.clone(BaseSpaceSource)
        yield self.clone(mergeFASTQs)
        yield self.clone(qualityCheck)
        yield self.clone(alignATACsamplesTask)
        yield self.clone(SamBedFilteringTask)
        yield self.clone(PeakCallingTask)



#!/usr/bin/env python
import json
from collections import OrderedDict

def argparser2dict(parser,arguments = None, indent = False):
    """
    Returns a dict with key prog name
    and value a list of the expected arguments.

    Subparser arg groups are in sublists and can thus be recognised when the
    value of a key is a list and not a dict.

    The dict should be transformable with json.dump[s]
    """
    import argparse
    if arguments is None:
        arguments = []
        returnFinalDict = True
    else: returnFinalDict = False
    for action in parser._actions:
        if type(action) is argparse._SubParsersAction:
            for subparser in action.choices:
                if subparser not in argparser2dict.filter:
                    arguments.append({'name':subparser,'type':'subparser','value':None,'args':[]})
                    argparser2dict(action.choices[subparser],arguments[-1]['args'])
        elif action.dest not in argparser2dict.filter:
            arg = vars(action).copy()
            arg['name'] = arg.pop('dest')
            arg['positional'] = False if arg['option_strings'] else True
            arg['type'] = argparser2dict.typeDict[arg.pop('type')]
            arg['choices'] = list(action.choices) if action.choices else None
            arg['value'] = None
            if arg['default'] is None: arg['default'] = ''
            arg.pop('container')
            arguments.append(arg)
    if returnFinalDict:
        return {parser.prog:arguments}
argparser2dict.typeDict = {
    bool: 'checkbox',
    int: 'number',
    float: 'number',
    str: 'text',
    None: 'checkbox'
    # store_true does not have a type set on the argument element (TODO documentation!), only for general genairics options not luigi derived args
}
argparser2dict.filter = {'help','console','server'}

def dict2argparsed(argparsedict,typecheck=False,plumbumCommand=True):
    """
    takes a arg dict and constructs the argparsed commandline 
    as a pumblum command ready for execution
    """
    from plumbum import local
    def recursive_args(arguments,cli_arguments):
        for arg in arguments:
            if arg['type'] == 'subparser':
                if arg['value'] == 'set':
                    cli_arguments.append(arg['name'])
                    recursive_args(arg['args'],cli_arguments)
            else:
                if arg['positional']:
                    cli_arguments.append(arg['value'])
                elif arg['const'] and arg['value']:
                    cli_arguments.append(arg['option_strings'][0])
                elif arg['value']: cli_arguments += [arg['option_strings'][0],arg['value']]
    arguments = argparsedict
    for prog in arguments:
        command = local[prog]
        cli_arguments = []
        recursive_args(arguments[prog],cli_arguments)
    if plumbumCommand:
        return command.bound_command(*cli_arguments)
    else: return cli_arguments

def webserver(parser,queue,jobstatus,stopevent):
    from flask import Flask, request, render_template
    app = Flask('genairics')

    @app.route('/')
    def offerJob():
        jobargs = argparser2dict(parser)
        return render_template('index.html',jobargs=jobargs)

    @app.route('/checkstatus',methods=['GET','POST'])
    def checkJobStatus():
        if request.json:
            data = request.json
            app.logger.debug("checking %s",data)
            try:
                return jobstatus[data["id"]]
            except KeyError:
                return "unknown (job submitted previously)"

    @app.route('/stopserver',methods=['GET'])
    def stopserver():
        stopevent.set()
        if queue.empty(): queue.put((None,None))
        return "Server closing down"
    
    @app.route('/submitjob',methods=['GET','POST'])
    def takeJSONjob():
        app.logger.info("genairics job submitted")
        app.logger.debug(request.json)
    
        if request.json:
            data = request.json
            jobargs = argparser2dict(parser)
            for prog in jobargs:
                for arg in jobargs[prog]:
                    if arg['type'] == 'subparser':
                        if arg['name'] == data['pipeline']:
                            arg['value'] = 'set'
                            for subarg in arg['args']:
                                try: subarg['value'] = data[subarg['name']]
                                except KeyError: app.logger.warning('pipeline option %s not set',subarg['name'])
                    else:
                        try: arg['value'] = data[arg['name']]
                        except KeyError: app.logger.warning('global option %s not set',arg['name'])
            command = dict2argparsed(jobargs)
            if queue: queue.put((data.get("jobid"),command))
            return "Job %s submitted" % data.get("jobid")

        else:
            return "no json job received"

    app.run()

def startJob(queue,jobstatus,stopevent):
    """
    Starts the jobs.
    To stop, stopevent needs to be set to true and
    an item needs to be put on the queue incase it is 
    empty, as the thread will otherwise hang waiting
    for another job.
    """
    from genairics import logger
    from genairics.__main__ import main
    while True:
        jobid,job = queue.get()
        if stopevent.is_set(): break
        jobstatus[jobid] = "started"
        stdout = job() # when the queue provides ready jobs
        logger.info(stdout)
        #main(job)
        jobstatus[jobid] = "finished"
        
def jobserver(args=None):
    import threading, queue, webbrowser, signal, sys
    from genairics import config, logger
    from genairics.__main__ import prepareParser
    logger.setLevel('DEBUG')

    parser = prepareParser()[0] #parser is first element of tuple returned

    stopevent = threading.Event()
    jobqueue = queue.Queue()
    jobstatus = {}
    threads = {
        'server': threading.Thread(target=webserver, daemon=True, args=(parser,jobqueue,jobstatus,stopevent)),
        'worker': threading.Thread(target=startJob, args=(jobqueue,jobstatus,stopevent))
    }
    for t in threads.values(): t.start()

    # open genairics wui in webbrowser
    if config.browser:
        webbrowser.get(config.browser).open('http://127.0.0.1:5000/',new=2)

    threads['worker'].join()
    #def signal_handler(signal, frame):
    #    print('You pressed Ctrl+C!')
    #    exit(0)
    #signal.signal(signal.SIGINT, signal_handler)
    #print('Press Ctrl+C to exit genairics server')
    #signal.pause()



from setuptools import setup

package = "genairics"
version = "0.1.31"

setup(name = package,
      version = version,
      description="GENeric AIRtight omICS pipelines",
      url='https://github.com/beukueb/genairics',
      author = 'Christophe Van Neste',
      author_email = 'christophe.vanneste@ugent.be',
      license = 'GNU GENERAL PUBLIC LICENSE',
      packages = ['genairics'],
      install_requires = [
          'luigi',
          'plumbum',
          'matplotlib',
          'seaborn',
          'pandas',
          'sklearn',
          'requests',
          'argcomplete',
      ],
      extras_require = {
          'interactive_console':  ["ipython"],
          'simulation': ["gffutils"],
          'ChIPseq': ["cutadapt"],
          'wui': ["Flask"]
      },
      package_data = {
          'genairics': [
              'scripts/genairics_dependencies.sh',
              'scripts/buildRSEMindex.sh',
	      'scripts/qualitycheck.sh',
              'scripts/simpleDEvoom.R',
              'templates/index.html',
              'static/js/main.js'
          ]
      },
      include_package_data = True,
      zip_safe = False,
      entry_points = {
          'console_scripts': [
              'genairics=genairics.__main__:main',
              'gaxs=genairics.utils:jobserver'
          ],
      },
      test_suite = 'nose.collector',
      tests_require = ['nose']
)

#To install with symlink, so that changes are immediately available:
#pip install -e .
